from yaml.representer import Representer
import collections
import logging
import yaml
import sys
import os

from bridgy import inventory
from bridgy.config.base import ConfigBase

logger = logging.getLogger()

CONFIG_TEMPLATE_NAME = 'sample_config_2.yml'


class Config(ConfigBase):

    def sources(self):
        sources = self.dig('inventory', 'source')
        ret = []
        for srcCfg in sources:
            source = srcCfg['type']
            ret.append( (source, srcCfg, ) )
        return ret

    @property
    def config_template_path(self):
        return CONFIG_TEMPLATE_NAME

    def verify(self):

        if self.dig('inventory', 'source') == None:
            logger.error("No inventory source specified (%s):" % self.path)
            sys.exit(1)
        
        if self.dig('inventory', 'include_pattern') != None and self.dig('inventory', 'exclude_pattern') != None:
            logger.error("'exclude_pattern' and 'include_pattern' are mutually exclusive")
            sys.exit(1)

        for source, srcCfg in self.sources():
            # verify each source here
            if source == 'newrelic':
                if 'insights_query_api_key' in srcCfg and srcCfg['insights_query_api_key'] == "API_KEY":
                    logger.error("New Relic inventory selected but no API key was specified: %s" % self.path)
                    sys.exit(1)



import os
import json
import requests
try:
    from urllib.parse import quote_plus
except ImportError:
    from urllib import quote_plus

from bridgy.inventory.source import InventorySource, Instance

class NewRelicInventory(InventorySource):

    name = 'newrelic'
    url = "https://insights-api.newrelic.com/v1/accounts/{}/query?nrql={}"

    def __init__(self, account_number, insights_query_api_key, data_path, proxies=None):
        self.account_number = account_number
        self.insights_query_api_key = insights_query_api_key
        self.data_file = os.path.join(data_path, '%s.json' % str(account_number))
        self.query = quote_plus("SELECT entityName, fullHostname, hostname, ipV4Address from NetworkSample LIMIT 999")
        if proxies:
            self.proxies = proxies
        else:
            self.proxies = {}

    def update(self):
        headers = {'X-Query-Key': self.insights_query_api_key,
                   'Accept': 'application/json'}

        response = requests.get(NewRelicInventory.url.format(self.account_number, self.query),
                                headers=headers,
                                proxies=self.proxies)

        with open(self.data_file, 'w') as data_file:
            data_file.write(response.text)

    def instances(self):
        instances = set()
        with open(self.data_file, 'r') as data_file:
            data = json.load(data_file)

        for results_dict in data['results']:
            for event_dict in results_dict['events']:
                hostname = event_dict['hostname']
                address = event_dict['ipV4Address'].strip().split("/")[0]
                if hostname == None:
                    hostname = address
                instances.add(Instance(hostname, address, None, self.name))

        return list(instances)



from yaml.representer import Representer
import collections
import logging
import yaml
import sys
import os

from bridgy import inventory
from bridgy.config.base import ConfigBase

logger = logging.getLogger()

CONFIG_TEMPLATE_NAME = 'sample_config_1.yml'


class Config(ConfigBase):

    def sources(self):
        source = self.dig('inventory', 'source')
        return list( (source, self[source],) )

    @property
    def config_template_path(self):
        return CONFIG_TEMPLATE_NAME

    def verify(self):
        source = self.dig('inventory', 'source')
        if source == None:
            logger.error("No inventory source specified (%s):" % self.path)
            sys.exit(1)

        if source == 'newrelic' and self.dig('newrelic', 'insights_query_api_key') == "API_KEY":
            logger.error("New Relic inventory selected but no API key was specified: %s" % self.path)
            sys.exit(1)

        if source not in self.conf.keys():
            logger.error("No inventory-specific section specified for %s source (%s):" % (repr(source), self.path))
            sys.exit(1)

        if self.dig('inventory', 'include_pattern') != None and self.dig('inventory', 'exclude_pattern') != None:
            logger.error("'exclude_pattern' and 'include_pattern' are mutually exclusive")
            sys.exit(1)



import collections
import logging
import pkgutil
import yaml
import abc
import sys
import os

from yaml.representer import Representer

from bridgy import inventory

logger = logging.getLogger()


class ConfigBase(object):
    __metaclass__ = abc.ABCMeta

    path = "~/.bridgy/config.yml"
    inventory = "~/.bridgy/inventory"
    mount = "~/.bridgy/mounts"
    conf = None

    def __init__(self, initial_data=None):
        # this is usefule for testing purposes
        self.conf = initial_data

    @abc.abstractproperty
    def config_template_path(self): pass

    @abc.abstractproperty
    def sources(self): pass

    @abc.abstractmethod
    def verify(self): pass

    @property
    def config_template_contents(self): 
        return pkgutil.get_data('bridgy', 'config/samples/' + self.config_template_path)

    def read(self):
        # ensure yaml uses a defaultdict(str)
        yaml.add_representer(collections.defaultdict,
                             Representer.represent_str)
        try:
            with open(os.path.expanduser(self.path), 'r') as fh:
                self.conf = yaml.load(fh)
        except Exception as ex:
            logger.error("Unable to read config (%s): %s" % (self.path, ex))
            sys.exit(1)

    def exists(self):
        config_file = os.path.expanduser(self.path)
        return os.path.exists(config_file)

    def create(self):
        if not self.exists():
            config_file = os.path.expanduser(self.path)
            logging.getLogger().info("Creating %s" % self.path)
            parent_dir = os.path.dirname(config_file)
            if not os.path.exists(parent_dir):
                os.mkdir(parent_dir)
            with open(config_file, 'w') as fh:
                fh.write(self.config_template_contents)

        inventory_cache = os.path.expanduser(self.inventory)
        if not os.path.exists(inventory_cache):
            parent_dir = os.path.dirname(inventory_cache)
            if not os.path.exists(parent_dir):
                os.mkdir(parent_dir)
            os.mkdir(inventory_cache)

        for source in list(inventory.SOURCES.keys()):
            source_path = os.path.join(inventory_cache, source)
            if not os.path.exists(source_path):
                os.mkdir(source_path)

        mount_path = os.path.expanduser(self.mount)
        if not os.path.exists(mount_path):
            os.mkdir(mount_path)

    def inventoryDir(self, source):
        if source not in list(inventory.SOURCES.keys()):
            raise RuntimeError(
                "Unexpected inventory source: %s" % repr(source))
        return os.path.join(os.path.expanduser(self.inventory),
                            source)

    @property
    def mount_root_dir(self):
        return os.path.expanduser(self.mount)

    def __iter__(self):
        return iter(self.conf)

    def dig(self, *original_keys):
        def __dig(d, *keys):
            try:
                if len(keys) == 1:
                    try:
                        return d[keys[0]]
                    except TypeError:
                        return None
                return __dig(d[keys[0]], *keys[1:])
            except KeyError:
                return None
        return __dig(self.conf, *original_keys)

    def __getitem__(self, key):
        return self.conf[key]

    def __setitem__(self, key, value):
        self.conf[key] = value

    def __repr__(self):
        return repr(self.conf)


import os
import shlex
import logging
import subprocess

logger = logging.getLogger()

def is_installed():
    if os.system('which tmux >/dev/null 2>&1') == 0:
        return True
    return False

def run(config, commands, in_windows=False, layout=None, dry_run=False, sync=False):
    layout_cmds = None
    if layout:
        layout_cmds = config.dig('tmux', 'layout', layout)
        if not layout_cmds:
            raise RuntimeError("Config does not define layout: %s" % layout)

    with TmuxSession(commands=commands, in_windows=in_windows, layout_cmds=layout_cmds, dry_run=dry_run, sync=sync) as tmux:
        tmux.attach()


# adapted from https://github.com/spappier/tmuxssh/
class TmuxSession(object):

    def __init__(self, session_name=None, commands=tuple(), in_windows=False, layout_cmds=None, dry_run=False, sync=False):
        self._session_name = session_name or 'tmux-{}'.format(os.getpid())
        self._commands = commands
        self._in_windows = in_windows
        self._layout_cmds = layout_cmds
        self._created_session = False
        self._show_errors = True
        self._dry_run = dry_run
        self._sync = sync

    def __enter__(self):
        if len(self._commands) == 0:
            return self

        # open a set of windows and run some commands
        if self._layout_cmds:
            for cmdIdx, (name, command) in enumerate(self._commands.items()):

                if cmdIdx == 0:
                    self.new_session(self._session_name, window_name=name, command=command)
                else:
                    # new window for all layout panes running the same cmd
                    self.new_window(name, command)

                # create each pane
                for idx, item in enumerate(self._layout_cmds):

                    cmd = [item['cmd'], '-t', name, command]
                    if 'run' in item:
                        cmd.append(item['run'])

                    self.tmux(*cmd)

                    # get rid of the first pane as this is not running a cmd
                    if idx == 0:
                        self.kill_pane(0)

                    self.select_layout('tiled')

        # open one window
        else:
            for cmdIdx, (name, command) in enumerate(self._commands.items()):
                if self._in_windows:
                    if cmdIdx == 0:
                        self.new_session(self._session_name, window_name=name, command=command)
                    else:
                        # new window for all layout panes running the same cmd
                        self.new_window(name, command)
                else:
                    if cmdIdx == 0:
                        self.new_session(self._session_name, window_name='remote-session', command=command)
                    else:
                        self.split_window(command)

                self.select_layout('tiled')

        if self._sync:
            self.set_window_option('synchronize-panes', 'on')

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.kill_session()

    def suppress_errors(func):
        def wrapper(self, *args, **kwargs):
            if self._created_session:
                old_show_errors = self._show_errors
                self._show_errors = False
                ret = func(self, *args, **kwargs)
                self._show_errors = old_show_errors
                return ret
        return wrapper

    def run_only_with_session(func):
        def wrapper(self, *args, **kwargs):
            if self._created_session:
                return func(self, *args, **kwargs)
        return wrapper

    def tmux(self, *args):
        cmd = ['tmux'] + list(args)
        logger.debug(' '.join(cmd))
        if self._dry_run:
            return ''

        pipes = subprocess.Popen(shlex.split(' '.join(cmd)), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        std_out, std_err = pipes.communicate()

        if pipes.returncode != 0 and self._show_errors:
            logger.error("Tmux failed (rc:%d, error:%s): %s" % (pipes.returncode,
                                                                repr(std_err.strip("\n").replace("\n",', ')),
                                                                " ".join(cmd)))

        elif len(std_err) and self._show_errors:
            logger.error("Tmux error: %s" % std_err)

        return std_out

    def new_session(self, session_name, window_name=None, command=None):
        cmd = ['new-session', '-ds', session_name]
        if window_name:
            cmd += ['-n', window_name]
        if command:
            cmd += [command]

        self.tmux(*cmd)
        self._created_session = True

    @run_only_with_session
    def new_window(self, name, command):
        if command:
            self.tmux('new-window', '-n', name, command)
        else:
            self.tmux('new-window', '-n', name)

    @run_only_with_session
    def split_window(self, command):
        self.tmux('split-window', '-t', self._session_name, command)

    @run_only_with_session
    def select_layout(self, layout):
        self.tmux('select-layout', '-t', self._session_name, layout)

    @run_only_with_session
    def attach(self):
        self.tmux('attach', '-t', self._session_name)

    @run_only_with_session
    def set_window_option(self, option, value):
        self.tmux('set-window-option', '-t', self._session_name, option, value)

    @run_only_with_session
    def kill_pane(self, n):
        self.tmux('kill-pane', '-t', str(n))

    @run_only_with_session
    def kill_pane(self, n):
        self.tmux('kill-pane', '-t', str(n))

    @suppress_errors
    @run_only_with_session
    def kill_session(self):
        self.tmux('kill-session', '-t', self._session_name)
        self._created_session = False



import os
import re
from functools import partial

from bridgy.utils import memoize
from bridgy.inventory.source import Instance, InventorySet
from bridgy.inventory.aws import AwsInventory
from bridgy.inventory.flatfile import CsvInventory
from bridgy.inventory.newrelic import NewRelicInventory
# from gcp import GcpInventory

SOURCES = {
    'aws': AwsInventory,
    # 'gcp': GcpInventory,
    'csv': CsvInventory,
    'newrelic': NewRelicInventory,
}

@memoize
def inventory(config):
    inventorySet = InventorySet()

    for source, srcCfg in config.sources():
        if source == 'aws':
            if os.path.exists(os.path.expanduser("~/.aws")):
                inventory = AwsInventory(cache_dir=config.inventoryDir(AwsInventory.name))
            else:
                inventory = AwsInventory(cache_dir=config.inventoryDir(AwsInventory.name),
                                         access_key_id=srcCfg['access_key_id'],
                                         secret_access_key=srcCfg['secret_access_key'],
                                         session_token=srcCfg['session_token'],
                                         region=srcCfg['region'])
            inventorySet.add(inventory)

        elif source == 'csv':
            csvPath = os.path.join(config.inventoryDir(source), srcCfg['name'])
            inventory = CsvInventory(path=csvPath,
                                     fields=srcCfg['fields'],
                                     delimiter=srcCfg['delimiter'] or ',' )
            inventorySet.add(inventory)

        elif source == 'newrelic':

            proxies = {}

            if config.dig('inventory', 'http_proxy'):
                proxies['http'] = config.dig('inventory', 'http_proxy')
            elif 'HTTP_PROXY' in os.environ:
                proxies['http'] = os.environ['HTTP_PROXY']
            elif 'http_proxy' in os.environ:
                proxies['http'] = os.environ['http_proxy']

            if config.dig('inventory', 'https_proxy'):
                proxies['https'] = config.dig('inventory', 'https_proxy')
            elif 'HTTPS_PROXY' in os.environ:
                proxies['https'] = os.environ['HTTPS_PROXY']
            elif 'https_proxy' in os.environ:
                proxies['https'] = os.environ['https_proxy']

            inventory = NewRelicInventory(account_number=srcCfg['account_number'],
                                          insights_query_api_key=srcCfg['insights_query_api_key'],
                                          data_path=config.inventoryDir(NewRelicInventory.name),
                                          proxies=proxies)
            inventorySet.add(inventory)
        
    return inventorySet

def instance_filter(instance, include_re=None, exclude_re=None):
    comparables = [instance.name, instance.address]

    if instance.aliases:
        comparables.extend(list(instance.aliases))

    if include_re:
        for name in comparables:
            if include_re.search(name):
                return True
        return False
    elif exclude_re:
        for name in comparables:
            if exclude_re.search(name):
                return False
        return True
    else:
        return True

@memoize
def instances(config):
    include_re, exclude_re = None, None
    if config.dig('inventory', 'include_pattern'):
        include_re = re.compile(config.dig('inventory', 'include_pattern'))
    if config.dig('inventory', 'exclude_pattern'):
        exclude_re = re.compile(config.dig('inventory', 'exclude_pattern'))

    all_instances = inventory(config).instances()
    config_instance_filter = partial(instance_filter, include_re=include_re, exclude_re=exclude_re)
    return list(filter(config_instance_filter, all_instances))

def search(config, targets):
    fuzzy = False
    if config.dig('inventory', 'fuzzy_search'):
        fuzzy = config.dig('inventory', 'fuzzy_search')

    include_re, exclude_re = None, None
    if config.dig('inventory', 'include_pattern'):
        include_re = re.compile(config.dig('inventory', 'include_pattern'))
    if config.dig('inventory', 'exclude_pattern'):
        exclude_re = re.compile(config.dig('inventory', 'exclude_pattern'))

    matched_instances = inventory(config).search(targets, fuzzy=fuzzy)
    config_instance_filter = partial(instance_filter, include_re=include_re, exclude_re=exclude_re)
    return list(filter(config_instance_filter, matched_instances))

def update(config):
    inventory(config).update()



import boto3
import placebo
import logging

from bridgy.inventory.source import InventorySource, Instance

logger = logging.getLogger()

class AwsInventory(InventorySource):

    name = 'aws'

    def __init__(self, cache_dir, access_key_id=None, secret_access_key=None, session_token=None, region=None):
        if access_key_id != None and secret_access_key != None and session_token != None and region != None:
            session = boto3.Session(
                aws_access_key_id=access_key_id,
                aws_secret_access_key=secret_access_key,
                aws_session_token=session_token,
                region_name=region
            )
        else:
            # pull from ~/.aws/* configs
            session = boto3.Session()

        self.pill = placebo.attach(session, data_path=cache_dir)

        self.client = session.client('ec2')

    def update(self):
        try:
            self.__ec2_search(stub=False)
        except KeyboardInterrupt:
            logger.error("Cancelled by user")

    def instances(self):
        data = self.__ec2_search(stub=True)

        instances = []
        for reservation in data['Reservations']:
            for instance in reservation['Instances']:

                # try to find the best dns/ip address to reach this box
                address = None
                if instance['PublicDnsName']:
                    address = instance['PublicDnsName']
                elif instance['PrivateIpAddress']:
                    address = instance['PrivateIpAddress']

                # try to find the best field to match a name against
                aliases = list()
                if 'Tags' in list(instance.keys()):
                    for tagDict in instance['Tags']:
                        if tagDict['Key'] == 'Name':
                            aliases.append(tagDict['Value'])
                            break

                if instance['PublicDnsName']:
                    aliases.append(instance['PublicDnsName'])
                if instance['PrivateDnsName']:
                    aliases.append(instance['PrivateDnsName'])
                if instance['InstanceId']:
                    aliases.append(instance['InstanceId'])

                aliases[:] = [x for x in aliases if x != None]
                name = aliases.pop(0)

                # take note of this instance
                if name != None and address != None:
                    if len(aliases) > 0:
                        instances.append(Instance(name, address, tuple(aliases), self.name))
                    else:
                        instances.append(Instance(name, address, self.name))

        return instances

    def __ec2_search(self, tag=None, value=None, stub=True):
        filters = []
        if value:
            filters.append({
                'Name': 'tag:' + tag,
                'Values': [value]
            })

        if stub:
            self.pill.playback()
            data = self.client.describe_instances(Filters=filters)
        else:
            self.pill.record()
            data = self.client.describe_instances(Filters=filters)
            self.pill.stop()
        return data



import os
import csv
import sys
import logging

from bridgy.inventory.source import InventorySource, Instance

logger = logging.getLogger()

class CsvInventory(InventorySource):

    name = 'csv'

    def __init__(self, path, fields, delimiter=','):
        self.csv_path = path
        self.fields = [f.strip() for f in fields.split(",")]
        self.delimiter = delimiter.strip()

    def update(self): pass

    def instances(self):
        instances = set()
        try:
            with open(self.csv_path, 'r') as csv_file:
                reader = csv.DictReader(csv_file, fieldnames=self.fields, delimiter=self.delimiter)
                for row in reader:
                    instances.add(Instance(row['name'].strip(), row['address'].strip(), None, self.name))
        except IOError as ex:
            logger.error("Unable to read inventory: %s" % ex)
            sys.exit(1)
        return list(instances)



import setuptools
import glob
import os

exec(open('./bridgy/version.py').read())

setuptools.setup(
    name='bridgy',
    version=__version__,
    url='https://github.com/wagoodman/bridgy',
    license=__license__,
    author=__author__,
    author_email=__email__,
    description='Easily search your cloud inventory and integrate with ssh + tmux + sshfs',
    packages=setuptools.find_packages(),
    package_data={
        'bridgy': ['config/samples/*.yml']
    },
    include_package_data=True,
    zip_safe=False,
    install_requires=['PyYAML',
                      'requests',
                      'docopt',
                      'inquirer',
                      'fuzzywuzzy',
                      'boto3',
                      'placebo',
                      'coloredlogs',
                      'tabulate',
                      'ansible',],
    platforms='linux',
    keywords=['tmux', 'ssh', 'sshfs', 'aws', 'newrelic', 'inventory', 'cloud'],
    # latest from https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers = [
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: System :: Monitoring',
        'Topic :: System :: Systems Administration',
        'Topic :: System :: System Shells',
        'Topic :: Terminals',
        'Topic :: Utilities',
        ],
    entry_points={
        'console_scripts': [
            'bridgy = bridgy.__main__:main'
        ]
    },
)



#!/usr/bin/env python
"""bridgy
SSH + TMUX + SSHFS + CLOUD INVENTORY SEARCH.
Fuzzy search for one or more systems then ssh into all matches, organized
by tmux.

Usage:
  bridgy ssh [-adsuvw] [-l LAYOUT] <host>...
  bridgy ssh [-dsuv] --tmux <host>
  bridgy list-inventory
  bridgy list-mounts
  bridgy mount [-duv] <host>:<remotedir>
  bridgy unmount [-dv] (-a | <host>...)
  bridgy run <task>
  bridgy update [-v]
  bridgy (-h | --help)
  bridgy --version

Sub-commands:
  ssh           ssh into the selected host(s)
  mount         use sshfs to mount a remote directory to an empty local directory
  unmount       unmount one or more host sshfs mounts
  list-mounts   show all sshfs mounts
  run           execute the given ansible task defined as playbook yml in ~/.bridgy/config.yml
  update        pull the latest inventory from your cloud provider

Options:
  -a        --all            Automatically use all matched hosts.
  -d        --dry-run        Show all commands that you would have run, but don't run them (implies --verbose).
  -l LAYOUT --layout LAYOUT  Use a configured lmux layout for each host.
  -s        --sync-panes     Synchronize input on all visible panes (tmux :setw synchronize-panes on).
  -t        --tmux           Open all ssh connections in a tmux session.
  -u        --update         pull the latest instance inventory from aws then run the specified command.
  -w        --windows        Use tmux windows instead of panes for each matched host.
  -h        --help           Show this screen.
  -v        --verbose        Show debug information.
  --version                  Show version.

Configuration Options are in ~/.bridgy/config.yml
"""
import sys
import os
import logging
import inquirer
import coloredlogs
import collections
from tabulate import tabulate
from docopt import docopt

from bridgy.version import __version__
from bridgy.command import Ssh, Sshfs, RunAnsiblePlaybook
import bridgy.inventory as inventory
import bridgy.config as cfg
import bridgy.tmux as tmux
import bridgy.utils as utils

logger = logging.getLogger()


def prompt_targets(question, targets=None, instances=None, multiple=True, config=None):
    if targets == None and instances == None or targets != None and instances != None:
        raise RuntimeError("Provide exactly one of either 'targets' or 'instances'")

    if targets:
        instances = inventory.search(config, targets)

    if len(instances) == 0:
        return []

    if len(instances) == 1:
        return instances

    display_instances = collections.OrderedDict()
    for instance in sorted(instances):
        display = "%-55s (%s)" % (instance.name, instance.address)
        display_instances[display] = instance

    questions = []

    if multiple:
        question = inquirer.Checkbox('instance',
                                     message="%s%s%s (space to multi-select, enter to finish)" % (utils.color.BOLD, question, utils.color.NORMAL),
                                     choices=list(display_instances.keys()) + ['all'],
                                     # default='all'
                                     )
    else:
        question = inquirer.List('instance',
                                 message="%s%s%s (enter to select)" % (utils.color.BOLD, question, utils.color.NORMAL),
                                 choices=list(display_instances.keys()),
                                 )
    questions.append(question)

    answers = None
    try:
        answers = inquirer.prompt(questions, raise_keyboard_interrupt=True)
    except KeyboardInterrupt:
        logger.error("Cancelled by user")
        sys.exit(1)

    if 'all' in answers["instance"]:
        selected_hosts = instances
    else:
        selected_hosts = []
        if not multiple:
            answers["instance"] = [answers["instance"]]
        for answer in answers["instance"]:
            selected_hosts.append(display_instances[answer])

    return selected_hosts


@utils.SupportedPlatforms('linux', 'windows', 'osx')
def ssh_handler(args, config):
    if config.dig('inventory', 'update_at_start') or args['-u']:
        update_handler(args, config)

    if args ['--tmux'] or config.dig('ssh', 'tmux'):
        question = "What instances would you like to ssh into?"
        targets = prompt_targets(question, targets=args['<host>'], config=config)
    else:
        question = "What instance would you like to ssh into?"
        targets = prompt_targets(question, targets=args['<host>'], config=config, multiple=False)

    if len(targets) == 0:
        logger.info("No matching instances found")
        sys.exit(1)

    commands = collections.OrderedDict()
    for idx, instance in enumerate(targets):
        name = '{}-{}'.format(instance.name, idx)
        commands[name] = Ssh(config, instance).command

    layout = None
    if args['--layout']:
        layout = args['--layout']

    if args['--tmux'] or config.dig('ssh', 'tmux'):
        tmux.run(config, commands, args['-w'], layout, args['-d'], args['-s'])
    else:
        cmd = list(commands.values())[0]
        if args['-d']:
            logger.debug(cmd)
        else:
            os.system(cmd)


@utils.SupportedPlatforms('linux', 'osx')
def mount_handler(args, config):
    Sshfs.ensure_sshfs_installed()

    if config.dig('inventory', 'update_at_start') or args['-u']:
        update_handler(args, config)

    fields = args['<host>:<remotedir>'].split(':')

    if len(fields) != 2:
        logger.error("Requires exactly 2 arguments: host:remotedir")
        sys.exit(1)

    desired_target, remotedir = fields
    instances = inventory.search(config, [desired_target])
    sshfs_objs = [Sshfs(config, instance, remotedir, dry_run=args['-d']) for instance in instances]
    unmounted_targets = [obj.instance for obj in sshfs_objs if not obj.is_mounted]

    question = "What instances would you like to have mounted?"
    target_instances = prompt_targets(question, instances=unmounted_targets, multiple=False, config=config)

    if len(target_instances) == 0:
        logger.info("No matching instances found")
        sys.exit(1)

    for sshfsObj in sshfs_objs:
        if sshfsObj.instance in target_instances:
            if sshfsObj.mount():
                logger.info("Mounted %s at %s" % (sshfsObj.instance.name, sshfsObj.mountpoint))
            else:
                logger.error("Unable to mount %s" % sshfsObj.instance.name)


@utils.SupportedPlatforms('linux', 'osx')
def list_mounts_handler(args, config):
    Sshfs.ensure_sshfs_installed()

    if args['-d']:
        return

    for mountpoint in Sshfs.mounts(config.mount_root_dir):
        logger.info(mountpoint)


@utils.SupportedPlatforms('linux', 'osx')
def unmount_handler(args, config):
    Sshfs.ensure_sshfs_installed()

    question = "What instances would you like to have unmounted?"

    if args['-a']:
        instances = inventory.instances(config)
        sshfs_objs = [Sshfs(config, instance, dry_run=args['-d']) for instance in instances]
        mounted_targets = [obj.instance for obj in sshfs_objs if obj.is_mounted]
        target_instances = mounted_targets
    else:
        desired_targets = args['<host>']
        instances = inventory.search(config, desired_targets)
        sshfs_objs = [Sshfs(config, instance, dry_run=args['-d']) for instance in instances]
        mounted_targets = [obj.instance for obj in sshfs_objs if obj.is_mounted]
        target_instances = prompt_targets(question, instances=mounted_targets, multiple=False, config=config)

    if len(target_instances) == 0:
        logger.error("No matching mounts found")
        if args['-a']:
            logger.warn("Did you select targets with <space> and confirm with <enter>?")
        sys.exit(1)

    for sshfsObj in sshfs_objs:
        if sshfsObj.instance in target_instances:
            if sshfsObj.unmount():
                logger.info("Unmounted %s" % sshfsObj.instance.name)
            else:
                logger.error("Unable to unmount %s" % sshfsObj.instance.name)


@utils.SupportedPlatforms('linux', 'windows', 'osx')
def list_inventory_handler(args, config):
    instances = []
    for ip, name, aliases, source in inventory.instances(config):
        if aliases:
            instances.append( (ip, name, ', '.join(aliases), source) )
        else:
            instances.append( (ip, name, '--- None ---', source) )
    logger.info(tabulate(instances, headers=['Name', 'Address/Dns', 'Aliases', 'Source']))


@utils.SupportedPlatforms('linux', 'windows', 'osx')
def update_handler(args, config):
    if args['-d']:
        return

    logger.warn("Updating inventory...")
    inventory_obj = inventory.inventory(config)
    inventory_obj.update()


@utils.SupportedPlatforms('linux', 'windows', 'osx')
def run_handler(args, config):
    # TODO: implement -d -a and -v

    task_name = args['<task>']
    task_playbook = config.dig('run', task_name)

    not_found = []
    desired_instances = []
    desired_targets = [ x.strip() for x in config.dig('run', task_name)[0]['hosts'].split(',') ]
    for desired_target in desired_targets:
        instances = inventory.search(config, [desired_target])
        if len(instances) == 0:
            not_found.append(desired_target)
            continue

        instance = inventory.Instance(desired_target, instances[0].address)
        desired_instances.append(instance)

    if len(not_found) > 0:
        logger.error("Unable to find instancesL %s" % ", ".join(not_found))
        sys.exit(1)

    if not task_playbook:
        logger.error("Playbook %s not configured." % repr(task_name))
        sys.exit(1)

    task = RunAnsiblePlaybook(task_name, task_playbook[0], config, desired_instances)
    task.run()


def main():
    coloredlogs.install(fmt='%(message)s')

    if os.geteuid() == 0:
        logger.error("Do not run this as root")
        sys.exit(1)

    config = cfg.Config()
    config.create()
    config.read()
    config.verify()

    version = 'bridgy %s' % __version__
    args = docopt(__doc__, version=version)

    if not tmux.is_installed():
        if args ['--tmux'] or config.dig('ssh', 'tmux'):
            logger.warn("Tmux not installed. Cannot support split screen.")
        args['--tmux'] = False

    if args['-v']:
        coloredlogs.install(fmt='%(message)s', level='DEBUG')

    if args['-d']:
        args['-v'] = True
        coloredlogs.install(fmt='%(message)s', level='DEBUG')
        logger.warn("Performing dry run, no actions will be taken.")

    if args['--version']:
        logger.info(version)
        sys.exit(0)

    opts = {
        'ssh': ssh_handler,
        'mount': mount_handler,
        'list-mounts': list_mounts_handler,
        'list-inventory': list_inventory_handler,
        'unmount': unmount_handler,
        'update': update_handler,
        'run': run_handler,
    }

    for opt, handler in list(opts.items()):
        if args[opt]:
            try:
                handler(args, config)
            except utils.UnsupportedPlatform as ex:
                logger.error(ex.message)
                sys.exit(1)

if __name__ == '__main__':
    main()



import logging
import sys
import os
import yaml
from collections import namedtuple
from tempfile import NamedTemporaryFile

from bridgy.command.error import *
from bridgy.command import Ssh

logger = logging.getLogger()

class RunAnsiblePlaybook(object):

    def __init__(self, name, playbook, config, instances):
        self.name = name
        self.playbook = playbook
        self.config = config
        self.instances = instances
        self.become_user = config.dig('ansible', 'become_user')
        self.become_method = config.dig('ansible', 'become_method')

    def _build_host_file_contents(self):
        inventory_str = ''
        for instance in self.instances:
            ssh_obj = Ssh(self.config, instance)
            instance_str = "{} ansible_host={}".format(instance.name, instance.address)

            options = ssh_obj.options
            if len(options) > 0:
                instance_str += " ansible_ssh_common_args=\"{}\"".format(options)

            user = self.config.dig('ssh', 'user')
            if user:
                instance_str += " ansible_user=\"{}\"".format(user)

            inventory_str += instance_str + '\n'
        return inventory_str

    def run(self):
        with NamedTemporaryFile(delete=True) as playbook_file:
            playbook_str = yaml.dump([self.playbook])
            playbook_file.write(playbook_str)
            playbook_file.flush()

            with NamedTemporaryFile(delete=True) as ansible_cfg:
                ansible_str = """
[defaults]
host_key_checking = False
stdout_callback = debug
remote_tmp = $HOME/.ansible/tmp
"""
                ansible_cfg.write(ansible_str)
                ansible_cfg.flush()

                with NamedTemporaryFile(delete=True) as hosts_file:

                    inventory_str = self._build_host_file_contents()
                    hosts_file.write(inventory_str)
                    hosts_file.flush()

                    options = {  
                                 'subset': None, 
                                 'ask_pass': False, 
                                 'listtags': None, 
                                 'become_method':  self.become_method or 'sudo', 
                                 'become_user':    self.become_user or 'root', 
                                 'sudo': False, 
                                 'private_key_file': None, 
                                 'syntax': None, 
                                 'skip_tags': [], 
                                 'diff': False, 
                                 'sftp_extra_args': '', 
                                 'check': False, 
                                 'force_handlers': False, 
                                 'remote_user': None, 
                                 'become_method': u'sudo', 
                                 'vault_password_file': None, 
                                 'listtasks': None, 
                                 'output_file': None, 
                                 'ask_su_pass': False, 
                                 'new_vault_password_file': None, 
                                 'inventory': u'hosts', 
                                 'forks': 100, 
                                 'listhosts': None, 
                                 'ssh_extra_args': '', 
                                 'tags': [u'all'], 
                                 'become_ask_pass': False, 
                                 'start_at_task': None, 
                                 'flush_cache': None, 
                                 'step': None, 
                                 'become': True, 
                                 'su_user': None, 
                                 'ask_sudo_pass': False, 
                                 'extra_vars': [], 
                                 'verbosity': 3, 
                                 'su': False, 
                                 'ssh_common_args': '', 
                                 'connection': 'ssh', 
                                 'ask_vault_pass': False, 
                                 'timeout': 10, 
                                 'module_path': None, 
                                 'sudo_user': None, 
                                 'scp_extra_args': ''
                                 }

                    os.environ['ANSIBLE_CONFIG'] = ansible_cfg.name

                    from ansible_utils import Runner

                    runner = Runner(
                        playbook=playbook_file.name,
                        hosts=hosts_file.name,
                        options=options
                    )

                    stats = runner.run()




from bridgy.command.error import *

class Ssh(object):

    def __init__(self, config, instance):
        if not hasattr(config, '__getitem__'):
            raise BadConfigError
        if not isinstance(instance, tuple):
            raise BadInstanceError

        self.config = config
        self.instance = instance

    @property
    def destination(self):
        if self.config.dig('ssh', 'user'):
            return '{user}@{host}'.format(user=self.config.dig('ssh', 'user'),
                                          host=self.instance.address)
        else:
            return self.instance.address

    @property
    def options(self):
        bastion = ''
        options = ''

        if 'bastion' in self.config:
            if not self.config.dig('bastion', 'address'):
                raise MissingBastionHost

            # build a destination from possible config combinations
            if self.config.dig('bastion', 'user'):
                destination = '{user}@{host}'.format(user=self.config.dig('bastion', 'user'),
                                                     host=self.config.dig('bastion', 'address'))
            else:
                destination = self.config.dig('bastion', 'address')

            bastion_options = self.config.dig('bastion', 'options') or ''

            template = "-o ProxyCommand='ssh {options} -W %h:%p {destination}'"
            bastion = template.format(options=bastion_options,
                                      destination=destination)

        options = self.config.dig('ssh', 'options') or ''

        return '{} {}'.format(bastion, options)


    @property
    def command(self):
        cmd = 'ssh {options} {destination}'
        return cmd.format(destination=self.destination,
                          options=self.options )
